(*
   BlockInspector.pas - Grid Block Inspector unit

   Copyright (C) 2013 - 2015 Martin Bittermann (martinbittermann@gmx.de)

   This file is part of ddrescueview.

   ddrescueview is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ddrescueview is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ddrescueview.  If not, see <http://www.gnu.org/licenses/>.
*)

unit BlockInspector;

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Grids, Spin, ExtCtrls, Shared, Parser, ImgBlockStore;

type

  { TBlockForm }

  TBlockForm = class(TForm, IFPObserver)
    BICol1Panel: TPanel;
    BICol2Panel: TPanel;
    BICol3Panel: TPanel;
    BlockEdit: TSpinEdit;
    BlockInfo: TGroupBox;
    EditBadSectors: TEdit;
    EditNonScrapedSectors: TEdit;
    EditNonTriedSectors: TEdit;
    EditRescuedSectors: TEdit;
    EditBlockInfo: TEdit;
    EditNonTrimmedSectors: TEdit;
    lblBad: TLabel;
    lblNonScraped: TLabel;
    lblNonTried: TLabel;
    lblNonTrimmed: TLabel;
    lblRescued: TLabel;
    BottomPanel: TPanel;
    TopPanel: TPanel;
    CloseBIButton: TButton;
    BlockHeadline: TLabel;
    LogEntryList: TStringGrid;
    TrackCurrentCheckBox: TCheckBox;
    procedure BlockEditChange(Sender: TObject);
    procedure LogEntryListHeaderSized(Sender: TObject; IsColumn: Boolean; Index: Integer);
    procedure CloseBIButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    FBlock : integer;
    FParser : TLogParser;
    FGridBlockStore : TImgBlockStore;
    procedure SetBlock(block : integer);
  public
    property Block : integer read FBlock write SetBlock;
    procedure DoUpdate(clearSel : boolean);
    Procedure FPOObservedChanged(ASender : TObject;
      Operation : TFPObservedOperation; Data : Pointer);
  end; 

var
  BlockForm: TBlockForm;

implementation
uses GUI, Math;

{$R *.lfm}

procedure TBlockForm.FormCreate(Sender: TObject);
begin
  FParser:=MainForm.Parser;
  FGridBlockStore:=MainForm.GridBlockStore;
  FGridBlockStore.AttachObserver(self);
end;

procedure TBlockForm.FormDestroy(Sender: TObject);
begin
  FGridBlockStore.DetachObserver(self);
  FGridBlockStore:=nil;
end;

Procedure TBlockForm.FPOObservedChanged(ASender : TObject;
  Operation : TFPObservedOperation; Data : Pointer);
begin
  if (ASender = FGridBlockStore) then begin
    DoUpdate(false);
  end;
end;

procedure TBlockForm.SetBlock(block : integer);
var
  activeblock, numBlocks : Longint;
begin
  activeblock := FGridBlockStore.activeBlock();
  numBlocks := FGridBlockStore.numBlocks;
  FBlock := max(0, min(block, numBlocks));
  // Disable Cur. Pos. tracking if requested block is not active block
  // or is outside of zooming bounds
  if (FBlock <> activeblock) or (activeblock < 0) or (activeBlock >= numBlocks) then
     TrackCurrentCheckBox.Checked:=false;
  // update the Form, clear the user's list selection
  DoUpdate(true);
end;

(* Updates the Block Inspector Form. If the TrackCurrentCheckBox is checked,
   it will change FBlock accordingly and display the current block instead.
   If clearSel is true, the selected row in the logEntry list is reset to
   the first item. *)
procedure TBlockForm.DoUpdate(clearSel : boolean);
var
  imgStartBlock, imgEndBlock, logEntry, logLen, row : Longint;
  bStart, bEnd, leStart, leEnd, overlap : Int64;
  rs : TRescueStatus; // only for counting sectors here
  BlockEditChangeEvent : TNotifyEvent;
begin
  with FGridBlockStore do begin
    // check if file is open
    if not Assigned(DataSource) or not DataSource.hasFile() then begin
      LogEntryList.RowCount:=1; // clear LogEntryList
      Hide;  // close BI window
      exit;
    end;

    // set the spinEdit bounds
    BlockEditChangeEvent:=BlockEdit.OnChange;
    BlockEdit.OnChange:=nil;
    BlockEdit.Value:=FBlock;
    BlockEdit.MaxValue:=numBlocks-1;
    BlockEdit.OnChange:=BlockEditChangeEvent;

    // adjust FBlock if curent pos tracking is enabled
    if TrackCurrentCheckBox.Checked then begin
      FBlock := activeBlock();
      BlockEdit.Value:=FBlock; // this also updates the MainForm
    end;


    rs:=emptyRescueStatus;
    bStart := getOffsetFromBlock(FBlock);
    bEnd := min(DataSource.rescueStatus.deviceSize,
                getOffsetFromBlock(block+1));

    EditBlockInfo.Text:='Start: ' + SizeStr(bStart) +
      '  Size: ' + SizeStr(bEnd-bStart) + ' (' +
      inttostr((bEnd-bStart) div devBSize) + ' sectors)';

    with DataSource do begin
      logLen := Length(log);
      // go through log and make list entries
      row := 0;
      for logEntry := 0 to logLen-1 do begin
        leStart:=log[logEntry].offset;
        leEnd:=leStart+log[logEntry].length;

        if leEnd <= bStart then continue;
        if leStart >= bEnd then break; // this and all following entries start after FBlock

        inc(row);
        if LogEntryList.RowCount <= row then LogEntryList.RowCount:=row+1; // need more space
        LogEntryList.Cells[0, row]:=BlockStatusToString(log[logEntry].status);
        // this log block is current rescue pos
        if (rescueStatus.pos >= leStart) and (rescueStatus.pos < leEnd) then
          LogEntryList.Cells[0, row]:='> '+LogEntryList.Cells[0, row];
        LogEntryList.Cells[1, row]:=inttostr(log[logEntry].offset)+' B ('+SizeStr(log[logEntry].offset)+')';
        LogEntryList.Cells[2, row]:=inttostr(log[logEntry].length)+' B ('+SizeStr(log[logEntry].length)+')';
        overlap:=BlockOverlap(leStart, leEnd, bStart, bEnd);
        case log[logEntry].status of
          '?' : inc(rs.nontried, overlap);
          '+' : inc(rs.rescued, overlap);
          '*' : inc(rs.nontrimmed, overlap);
          '/' : inc(rs.nonscraped, overlap);
          '-' : inc(rs.bad, overlap);
        end;
      end;
      LogEntryList.RowCount:=row+1;
      if clearSel then LogEntryList.Row:=1;
    end; // with FParser
    EditNontriedSectors.Text:=inttostr(rs.nontried div FGridBlockStore.devBSize);
    EditRescuedSectors.Text:=inttostr(rs.rescued div FGridBlockStore.devBSize);
    EditNonTrimmedSectors.Text:=inttostr(rs.nontrimmed div FGridBlockStore.devBSize);
    EditNonScrapedSectors.Text:=inttostr(rs.nonscraped div FGridBlockStore.devBSize);
    EditBadSectors.Text:=inttostr(rs.bad div FGridBlockStore.devBSize);
  end; // with FGridBlockStore
end;

// user closes the form
procedure TBlockForm.CloseBIButtonClick(Sender: TObject);
begin
  Hide;
end;

// The block number in the spind-edit has been changed
procedure TBlockForm.BlockEditChange(Sender: TObject);
begin
  if BlockEdit.Value <> FGridBlockStore.activeBlock then TrackCurrentCheckBox.Checked:=false;
  MainForm.SelectBlock(BlockEdit.Value);
  SetBlock(BlockEdit.Value);
end;

// prevent 'Status' column from becoming invisible
procedure TBlockForm.LogEntryListHeaderSized(Sender: TObject; IsColumn: Boolean;
  Index: Integer);
begin
  if LogEntryList.Columns.Items[0].Width < LogEntryList.Columns.Items[0].MinSize then
     LogEntryList.Columns.Items[0].Width := LogEntryList.Columns.Items[0].MinSize;
end;


end.

