(*
   Parser.pas - Log file parser unit

   Copyright (C) 2013 - 2015 Martin Bittermann (martinbittermann@gmx.de)

   This file is part of ddrescueview.

   ddrescueview is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ddrescueview is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ddrescueview.  If not, see <http://www.gnu.org/licenses/>.
*)

unit Parser;

interface
uses Classes, Shared;

type

  { TLogParser }

  TLogParser = class(TObservablePersistent)
  private
    FLog : TLog;
    FRescueStatus : TRescueStatus;
    FComments : TStringList; // comment lines of logfile
    FLogStream : TStream;
    FFileName : String;
    procedure logMsg(msg : String);
    function isEmptyLine(strLine : String) : boolean;
    function isCommentLine(strLine : String) : boolean;
  public
    destructor Destroy; override;
    procedure OpenFile(filename : String);
    procedure CloseFile;
    procedure parse();
    function hasFile() : boolean;
    property log : TLog read FLog;
    property CommentLines : TStringList read FComments;
    property rescueStatus : TRescueStatus read FRescueStatus;
    property LogFileName :String read FFileName;
  end;


implementation
uses SysUtils, GUI;

destructor TLogParser.Destroy;
begin
  CloseFile();
  inherited;
end;

procedure TLogParser.OpenFile(filename : String);
begin
  try
    CloseFile(); // close already open file
    // open the log file using shared read access, so ddrescue can still write to it.
    FLogStream := TFileStream.Create(filename, fmOpenRead or fmShareDenyNone);
    FComments := TStringList.Create;
    FFileName := filename;
    parse();  // start parsing after opening
  except
    on E: Exception do begin
        CloseFile;
        logMsg('Error: '+E.Message);
    end;
  end;
end;

procedure TLogParser.CloseFile;
begin
  FFileName := '';
  FreeAndNil(FComments);
  FreeAndNil(FLogStream);
  FRescueStatus := emptyRescueStatus;
  SetLength(FLog, 0);
  FPONotifyObservers(self, ooDeleteItem, nil);
end;

(* returns true if a line is empty *)
function TLogParser.isEmptyLine(strLine : String) : boolean;
begin
  strLine := TrimLeft(strLine);
  result := (Length(strLine) = 0); // empty line
end;

(* returns true if a line is a comment line *)
function TLogParser.isCommentLine(strLine : String) : boolean;
begin
  result := (strLine[1] = '#'); // comment line
end;


(* add a log line to the application log *)
procedure TLogParser.logMsg(msg : String);
begin
  MainForm.DbgLog.Lines.Add(msg);
end;

(* Parse a log file.
   After parsing, the results are available in
   the properties log and rescueStatus *)
procedure TLogParser.parse();
var
  line : string;
  token : array[0..2] of string;
  i, logEntry, lineIdx, idx : Integer;
  logStrings : TStringList;
  comments_end : boolean;
begin
  if not Assigned(FLogStream) then begin
    logMsg('Parser: No log file opened.');
    FPONotifyObservers(self, ooChange, nil);
    exit;
  end;

  try
    FLogStream.Seek(0, soFromBeginning);
    logStrings := TStringList.Create;
    logStrings.LoadFromStream(FLogStream);
    logMsg('Parsing log file: ' +IntToStr(logStrings.Count) + ' lines.');
  except
    on E : Exception do begin
      logMsg('Error: Cannot read log file: '+E.Message+'('+E.ClassName+')');
      FPONotifyObservers(self, ooChange, nil);
      exit;
    end;
  end;

  FRescueStatus := emptyRescueStatus;
  lineIdx := -1;
  logEntry := 0;
  comments_end := false;

  try
    repeat
      inc(lineIdx);
      line:=logStrings[lineIdx];
      if isCommentLine(line) then begin
        if comments_end or (pos('# current_pos', line) > 0) then
          comments_end:=true
        else FComments.Add(line);
        // process comment info
        if pos('Command line:', line) > 0 then begin
           repeat // not a loop, but goto replacement
               // try to find ddrescue's block size argument
               idx := pos(' -b', line);
               if idx <> 0 then idx := idx+3 // point to start of number
               else begin
                 idx := pos(' --block-size=', line);
                 if idx <> 0 then idx := idx+14 // point to start of number
                 else break; // argument not found, jump over the rest
               end;
               token[0] :=TrimLeft(Copy(line, idx, Length(line)));
               idx := pos(' ', token[0]); //space after argument
               if idx <> 0 then token[0] :=Copy(token[0], 1, idx-1);
               FRescueStatus.suggestedBlockSize:=StrToIntDef(token[0], 0);
           until true;
        end;
      end else if not isEmptyLine(line) then begin
        // split line into maximum of 3 tokens
        line:=Trim(line);
        for i:= 0 to 2 do begin
          idx:= Pos(' ', line);
          if idx = 0 then begin
            token[i] := line;
            line := '';
          end else begin
            token[i] := Copy(line, 1, idx-1);
            line := TrimLeft(Copy(line, idx, Length(line)-idx+1));
          end;
        end;
        if (token[0] <> '') and (token[1] <> '') then begin
          if token[2] <> '' then begin // standard block line (3 tokens)
            if Length(FLog) = logEntry then SetLength(FLog, logEntry+256); // allocate more space
            FLog[logEntry].offset:=StrToInt64(token[0]);
            FLog[logEntry].length:=StrToInt64(token[1]);
            FLog[logEntry].status:=token[2][1];
            case FLog[logEntry].status of
              '?' : inc(FRescueStatus.nontried, FLog[logEntry].length);
              '+' : inc(FRescueStatus.rescued, FLog[logEntry].length);
              '*' : inc(FRescueStatus.nontrimmed, FLog[logEntry].length);
              '/' : inc(FRescueStatus.nonscraped, FLog[logEntry].length);
              '-' : inc(FRescueStatus.bad, FLog[logEntry].length);
            end;
            if FLog[logEntry].status in ['-', '*', '/'] then inc(FRescueStatus.errors);
            inc(logEntry);
          end else begin // found the status line (2 components)
            if FRescueStatus.pos = 0 then begin
              FRescueStatus.pos:=StrToInt64(token[0]);
              FRescueStatus.curOperation:=token[1][1];
              FRescueStatus.strCurOperation:=OperationToText(FRescueStatus.curOperation);
            end else begin
              logMsg('Parser: found more than one line with 2 tokens.');
            end;
          end;
        end;
      end;
    until lineIdx = logStrings.Count-1;
    SetLength(FLog, logEntry); // trim array to actually needed size
  except
    on E : Exception do logMsg('Error parsing log file: '+E.Message+'('+E.ClassName+')');
  end;
  FreeAndNil(logStrings);

  if logEntry > 0 then begin
    // calculate device's size from last block's offset and length
    FRescueStatus.devicesize:=FLog[logEntry-1].offset + FLog[logEntry-1].length;
  end else logMsg('Parser: No blocks in logfile!');
  // notify any observers of the changes
  FPONotifyObservers(self, ooChange, nil);
end;

function TLogParser.hasFile: boolean;
begin
  hasFile:=Length(Flog)<>0;
end;

end.
