(*
   Shared.pas - Shared functionality

   Copyright (C) 2013 - 2015 Martin Bittermann (martinbittermann@gmx.de)

   This file is part of ddrescueview.

   ddrescueview is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ddrescueview is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ddrescueview.  If not, see <http://www.gnu.org/licenses/>.
*)

unit Shared;

(* This unit contains shared functionality, especially constants,
   types and essential functions for multiple parts of the program *)

interface

uses Classes, Graphics;

type
  PTextFile = ^TextFile;
  TLogEntry = record
    offset : int64;
    length : int64;
    status : char;
  end;
  TLog = Array of TLogEntry;
  TRescueStatus = record
    devicesize : int64;
    suggestedBlockSize : integer;
    pos : int64;
    rescued : int64;
    nontried : int64;
    bad : int64;
    nonscraped : int64;
    nontrimmed : int64;
    errors : int64;
    curOperation : Char;
    strCurOperation : String;
  end;
  IAttachableObserved = Interface(IFPObserved)
    (* Extend the IFPObserved intf to have Attach/detach wrappers
       that can be implemented as public methods by implementing classes.
       This would make these classes actually 'attachable' from the outside. *)
    procedure AttachObserver(AObserver : TObject);
    procedure DetachObserver(AObserver : TObject);
  end;

  TObservablePersistent = class(TPersistent, IAttachableObserved)
  (* Implement the IAttachableObserved interface for TPersistent.
     Predefined TPersistent descendants that need to be attachable should
     implement it like here *)
  public
    procedure AttachObserver(AObserver : TObject);
    procedure DetachObserver(AObserver : TObject);
  end;


const
  // block type masks for each block flag.
  // The upper 8 bits are reserved for these flags,
  // the lower 24 bits contain the color used for display.
  // The actual display color is blended together from the flags
  // in the upper 8 bits by ColorizeBlockMask
  MASK_NON_TRIED = $01000000;
  MASK_NON_TRIMMED = $02000000;
  MASK_NON_SCRAPED = $04000000;
  MASK_BAD_SECT = $08000000;
  MASK_FINISHED = $10000000;
  MASK_ALL_STATUSES = $1F000000;
  MASK_ACTIVE = $20000000;
  MASKS : array[0..4] of Longint =
    (MASK_NON_TRIED, MASK_FINISHED, MASK_NON_TRIMMED, MASK_NON_SCRAPED, MASK_BAD_SECT);

  // color defaults for the block statuses
  DEF_COLOR_NON_TRIED = $909090;
  DEF_COLOR_NON_TRIMMED = $00e0ff;
  DEF_COLOR_NON_SCRAPED = $ff2020;
  DEF_COLOR_BAD_SECT = $0000ff;
  DEF_COLOR_FINISHED = $20e020;
  DEF_COLOR_UNDEFINED = $606060;
  DEF_COLOR_ACTIVE = $ffff00;

  // color weight defaults to be used by ColorizeBlockMask for blending
  DEF_WEIGHT_NON_TRIED = 1;
  DEF_WEIGHT_FINISHED = 2;
  DEF_WEIGHT_NON_TRIMMED = 4;
  DEF_WEIGHT_NON_SCRAPED = 10;
  DEF_WEIGHT_BAD_SECT = 40;

  // color primary masks
  MASK_B = $ff0000;
  MASK_G = $00ff00;
  MASK_R = $0000ff;

  // default device block size
  DEF_BSIZE = 512;

  PROGRAM_TITLE = 'ddrescue log viewer';
  VERSION_MAJOR = '0';
  VERSION_MINOR = '4 alpha';
  emptyRescueStatus : TRescueStatus =
  (devicesize : 0; suggestedBlockSize : DEF_BSIZE; pos : 0; rescued : 0; nontried : 0; bad : 0;
   nonscraped : 0; nontrimmed : 0; errors : 0; curOperation : #0; strCurOperation : '');

var
  useDecimalUnits : boolean = true;
  // colors for the block statuses
  COLOR_NON_TRIED : Longint = $909090;
  COLOR_NON_TRIMMED : Longint = $00e0ff;
  COLOR_NON_SCRAPED : Longint = $ff2020;
  COLOR_BAD_SECT : Longint = $0000ff;
  COLOR_FINISHED : Longint = $20e020;
  COLOR_UNDEFINED : Longint = $606060;
  COLOR_ACTIVE : Longint = $ffff00;
  // color weights to be used by ColorizeBlockMask for blending
  WEIGHT_NON_TRIED : Longint = 1;
  WEIGHT_FINISHED : Longint = 2;
  WEIGHT_NON_TRIMMED : Longint = 4;
  WEIGHT_NON_SCRAPED : Longint = 10;
  WEIGHT_BAD_SECT : Longint = 40;
  // arrays containing the above constants for easier looping
  COLORS : array[0..4] of ^Longint =
    (@COLOR_NON_TRIED, @COLOR_FINISHED, @COLOR_NON_TRIMMED, @COLOR_NON_SCRAPED, @COLOR_BAD_SECT);
  WEIGHTS : array[0..4] of ^Longint =
    (@WEIGHT_NON_TRIED, @WEIGHT_FINISHED, @WEIGHT_NON_TRIMMED, @WEIGHT_NON_SCRAPED, @WEIGHT_BAD_SECT);

function BlockStatusToString(status: char) : String;
function OperationToText(status: char) : string;
function SizeStr(sizeInBytes : int64): String;
function BlockOverlap(b1Start, b1End, b2Start, b2End : Int64): Int64;


implementation
uses SysUtils, Math;


function BlockStatusToString(status: char) : String;
begin
  case status of
    '?': result := 'Non-tried';
    '*': result := 'Non-trimmed';
    '/': result := 'Non-scraped';
    '-': result := 'Bad sector(s)';
    '+': result := 'Rescued';
  else
    result := 'Unknown status';
  end;
end;

// status strings for the status line
function OperationToText(status: char) : string;
begin
  case status of
    '?': OperationToText := 'Copying non-tried';
    '*': OperationToText := 'Trimming non-trimmed blocks';
    '/': OperationToText := 'Scraping non-scraped blocks';
    '-': OperationToText := 'Retrying bad sectors';
    'F': OperationToText := 'Filling specified blocks';
    'G': OperationToText := 'Generating approximate logfile';
    '+': OperationToText := 'Finished';
  else
    OperationToText := 'Unknown operation';
  end;
end;

function SizeStr(sizeInBytes : int64): String;
begin
  if useDecimalUnits then begin
     if sizeInBytes < 0 then SizeStr := 'invalid size?'
     else if sizeInBytes < 100000 then SizeStr := IntToStr(sizeInBytes)+ ' Byte'
     else if sizeInBytes < 100000000 then SizeStr := IntToStr(sizeInBytes div 1000)+ ' KB'
     else if sizeInBytes < 100000000000 then SizeStr := IntToStr(sizeInBytes div 1000000)+ ' MB'
     else if sizeInBytes < 100000000000000 then SizeStr := IntToStr(sizeInBytes div 1000000000)+ ' GB'
     else if sizeInBytes < 100000000000000000 then SizeStr := IntToStr(sizeInBytes div 1000000000000)+ ' TB'
     else SizeStr := IntToStr(sizeInBytes div 1000000000000000)+ ' PB';
  end else begin
     if sizeInBytes < 0 then SizeStr := 'invalid size?'
     else if sizeInBytes < 102400 then SizeStr := IntToStr(sizeInBytes)+ ' Byte'
     else if sizeInBytes < 104857600 then SizeStr := IntToStr(sizeInBytes div 1024)+ ' KiB'
     else if sizeInBytes < 107374182400 then SizeStr := IntToStr(sizeInBytes div 1048576)+ ' MiB'
     else if sizeInBytes < 109951162777600 then SizeStr := IntToStr(sizeInBytes div 1073741824)+ ' GiB'
     else if sizeInBytes < 112589990684262400 then SizeStr := IntToStr(sizeInBytes div 1099511627776)+ ' TiB'
     else SizeStr := IntToStr(sizeInBytes div 1125899906842624)+ ' PiB';
  end;
end;

function BlockOverlap(b1Start, b1End, b2Start, b2End: Int64): Int64;
begin
  BlockOverlap:=Max(0, Min(Min(b1End-b1Start, b2End-b2Start),
                           Min(b2End-b1Start, b1End-b2Start)));
end;

procedure TObservablePersistent.AttachObserver(AObserver : TObject);
begin
  FPOAttachObserver(AObserver);
end;

procedure TObservablePersistent.DetachObserver(AObserver : TObject);
begin
  FPODetachObserver(AObserver);
end;

end.
