(*
   ImgBlockStore.pas - Image Block Storage unit

   Copyright (C) 2013 - 2015 Martin Bittermann (martinbittermann@gmx.de)

   This file is part of ddrescueview.

   ddrescueview is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ddrescueview is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ddrescueview.  If not, see <http://www.gnu.org/licenses/>.
*)

unit ImgBlockStore;

interface

uses
  Classes, SysUtils, Parser, Shared;

type

  { TImgBlockStore }

  TImgBlockStore = class(TObservablePersistent, IFPObserver)
  private
    FImgBlocks: Array of Longword; // lower 24 bits : color, upper 8 bits: status mask
    FDataSource : TLogParser; // parser / log source to use
    FNumBlocks : Longint;  // number of blocks requested
    FDevBSize : Int64; // device block size
    FBOffset : Int64; // imgBlocks start after this # of device sectors (zooming)
    FBCount : Int64; // number of device sectors to show in imgBlocks (zooming)
    FSectorsPerBlock : Int64; // device sectors per imgBlock
    FBytesPerBlock : Int64; // #of bytes in every imgBlock, multiple of FDevBSize
    FOwner : TObject;
    procedure assignDataSource(source : TLogParser);
    procedure buildBlocks;
    function BlockStatusToMask(status: char) : Longword;
    function ColorizeBlockMask(mask: Longint) : Longint;
    function getBlock(i: Longint): Longword;
    procedure setDevBSize(size : Int64);
  public
    constructor Create(AOwner : TObject); virtual;
    destructor Destroy; override;
    property Blocks[i:Longint] : Longword read getBlock; default;
    property DataSource : TLogParser read FDataSource write assignDataSource;
    property numBlocks : Longint read FNumBlocks write FNumBlocks;
    property devBSize : Int64 read FDevBSize write setDevBSize;
    property BlockOffset : Int64 read FBOffset write FBOffset;
    property BlockCount : Int64 read FBCount write FBCount;
    property bytesPerBlock : Int64 read FBytesPerBlock;
    function getOffsetFromBlock(block : Longint) : Int64;
    function getBlockFromOffset(offset : Int64) : Longint;
    function activeBlock() : integer;
    Procedure zoom(block: Longint; factor: double);
    Procedure FPOObservedChanged(ASender : TObject; //for the Observer role
      Operation : TFPObservedOperation; Data : Pointer);
  end;

implementation
uses GUI, Math;

// creates the ImgBlockStore and sets up listening for the owner.
constructor TImgBlockStore.Create(AOwner : TObject);
begin
  inherited Create;
  FDevBSize := DEF_BSIZE;
  FBytesPerBlock := DEF_BSIZE;
  FSectorsPerBlock := 1;
  FOwner := AOwner;
  if (AOwner<>nil) and (AOwner is IAttachableObserved) then
    (AOwner as IAttachableObserved).AttachObserver(self); // listen for requests from the owner
end;

destructor TImgBlockStore.Destroy;
begin
  SetLength(FImgBlocks, 0);
  if FDataSource<>nil then FDataSource.DetachObserver(self);
  if (FOwner<>nil) and (FOwner is IAttachableObserved) then
    (FOwner as IAttachableObserved).DetachObserver(self);
  inherited;
end;

procedure TImgBlockStore.FPOObservedChanged(ASender: TObject;
  Operation: TFPObservedOperation; Data: Pointer);
begin
  if ASender = FDataSource then begin // parser did its thing ...
    if Operation in [ooAddItem, ooDeleteItem] then begin // new or closed file!
      FDevBSize:=FDataSource.rescueStatus.suggestedBlockSize;
      FBOffset:=0; // reset offset
      FBCount:=0; // and # of displayed blocks (zoom)
    end;
    buildBlocks;
    FPONotifyObservers(self, ooChange, nil);
  end else begin // someone else sent an update request
    buildBlocks;
    FPONotifyObservers(self, ooChange, nil);
  end;
end;

// Assign a LogParser to the img Blocks.
// Assign nil before freeing the LogParser from the outside!
procedure TImgBlockStore.assignDataSource(source : TLogParser);
begin
  if FDataSource<>nil then FDataSource.DetachObserver(self); // detach old source
  FDataSource:=source;
  FDevBSize:=FDataSource.rescueStatus.suggestedBlockSize;
  FBOffset:=0; // reset offset
  FBCount:=0; // and # of displayed blocks (zoom)
  FDataSource.AttachObserver(self);
  buildBlocks;
  FPONotifyObservers(self, ooChange, nil);
end;

function TImgBlockStore.getBlock(i: Longint): Longword;
begin
  getBlock:=FImgBlocks[i];
end;

// Gets the Offset of a given imgBlock on the rescue device, in bytes.
function TImgBlockStore.getOffsetFromBlock(block: Longint): Int64;
begin
  getOffsetFromBlock:=FBOffset*FDevBSize + FBytesPerBlock*block;
end;

// Gets the block index that belongs to a given offset on the device.
// Aside from round, this is the opposite of getOffsetFromBlock.
// Can be negative or greater than numBlocks-1 when zoomed.
function TImgBlockStore.getBlockFromOffset(offset : Int64) : Longint;
var localOffset : Int64;
begin
  localOffset := offset - FBOffset*FDevBSize;
  getBlockFromOffset:=(localOffset div FBytesPerBlock) -
    Int64((localOffset mod FBytesPerBlock <> 0) and (localOffset < 0));
end;

// returns imgblock index of current position. May be out of range, see above.
function TImgBlockStore.activeBlock() : integer;
begin
  if Assigned(FDataSource) then
    result:=getBlockFromOffset(FDataSource.rescueStatus.pos)
  else result:=0;
end;


(* Calculate image blocks from log for fast drawing.
   Blocks will be stored in the imgBlocks[] array.
   Blocks contain type information and a color derived from the mix of types.
   Each block spans a multiple of the device sector size.
   This can lead to some unused blocks at the end of the grid, especially with
   small rescue devices, high number of imgBlocks or big sector size *)
procedure TImgBlockStore.buildBlocks;
var logLen, logBlock, imgBlock, imgStartBlock, imgEndBlock: integer;
begin
  if not Assigned(FDataSource) then logLen := 0
  else logLen := Length(FDataSource.log);

  // adjust block array to requested size (FNumBlocks)
  if Length(FImgBlocks) <> FNumBlocks then SetLength(FImgBlocks, FNumBlocks);

  if FNumBlocks=0 then exit;

  if (logLen = 0) then begin // empty log: un-zoom, gray out all blocks and exit
    FBOffset:=0;
    FBCount:=0;
    FillDword(FImgBlocks[0], Length(FImgBlocks), COLOR_NON_TRIED);
    exit;
  end else FillDword(FImgBlocks[0], Length(FImgBlocks), $00000000);//zero-out all blocks

  with FDataSource do begin
    // if not yet set, have zoom span whole drive
    if FBCount=0 then begin
       FBOffset:=0;
       FBCount:=FDataSource.rescueStatus.devicesize div FDevBSize;
    end;
    (* Calculate the number of bytes per image block. This is basically
     FBCount / FNumBlocks * FDevBSize with some more magic to ensure that all
     device sectors can be shown in the block array. The result is always a
     multiple of FDevBSize. Then sanitize FBCount accordingly.
     *)
    FSectorsPerBlock := (FBCount div FNumBlocks + Int64((FBCount mod FNumBlocks)<>0));
    FBytesPerBlock := FSectorsPerBlock * FDevBSize;
    FBCount := Min(FSectorsPerBlock*FNumBlocks,
                   FDataSource.rescueStatus.devicesize div FDevBSize - FBOffset);
    // iterate over log and collect block info
    for logBlock := 0 to logLen-1 do begin
      // skip blocks before zoomed-in area
      if (log[logBlock].offset+log[logBlock].length) <= FBOffset*FDevBSize then Continue;
      // skip blocks after zoomed-in area
      if (log[logBlock].offset) >= (FBCount+FBOffset)*FDevBSize then Break;
      // first img block which straddles current log entry
      imgStartBlock := (log[logBlock].offset-FBOffset*FDevBSize) div FBytesPerBlock;
      imgEndBlock := (log[logBlock].offset+log[logBlock].length-FBOffset*FDevBSize-1) div FBytesPerBlock;
      for imgBlock := Max(0,imgStartBlock) to Min(FNumBlocks-1, imgEndBlock) do begin
        FImgBlocks[imgBlock]:=FImgBlocks[imgBlock] or BlockStatusToMask(log[logBlock].status);
      end;
    end;

    // mark active block, i.e. block of current input read position
    if (rescueStatus.pos >= FBOffset*FDevBSize) and
       (rescueStatus.pos < (FBOffset+FBCount)*FDevBSize) then begin
      imgBlock := (rescueStatus.pos-FBOffset*FDevBSize) div FBytesPerBlock;
      FImgBlocks[imgBlock]:=FImgBlocks[imgBlock] or MASK_ACTIVE;
    end;

    // go through blocks and build colors
    for imgBlock := 0 to FNumBlocks-1 do begin
      FImgBlocks[imgBlock]:=ColorizeBlockMask(FImgBlocks[imgBlock]);
    end;
  end;
end;

(* computes a color for a block, depending on whick block types exist in
   this block. The percentage of sectors of a specific block type has no influence.
   The weight of a color (block type) can be set by the user, but is at default:
   (least) non-tried, finished, non-trimmed, non-scraped, bad (highest)
   Each of these types influences the block color depending on its weight.
   This way, small local quantities of important (e.g. bad) blocks are easy
   to notice over less important (rescued, non-tried) ones.
   TODO: (1) Optimize this for speed! Eliminate ifs and divs! *)
function TImgBlockStore.ColorizeBlockMask(mask: Longint) : Longint;
var i, b, g, r, totalweight, statusPresent : Longint;
begin
  b:=0; g:=0; r:=0; totalweight:=0;
  if not LongBool(mask and MASK_ALL_STATUSES) then begin
    result := mask or COLOR_UNDEFINED;
    exit;
  end;
  for i:= 0 to Length(MASKS)-1 do begin
    statusPresent:=Longint((mask and MASKS[i]) <> 0);
    b := b + statusPresent*WEIGHTS[i]^*((COLORS[i]^ and MASK_B) shr 16);
    g := g + statusPresent*WEIGHTS[i]^*((COLORS[i]^ and MASK_G) shr 8);
    r := r + statusPresent*WEIGHTS[i]^* (COLORS[i]^ and MASK_R);
    inc(totalweight, statusPresent*WEIGHTS[i]^);
  end;
  b := b div totalweight;
  g := g div totalweight;
  r := r div totalweight;
  result := mask or (b shl 16) or (g shl 8) or r;
end;

(* sets the device sector size and tries to maintain zoom as much as possible *)
procedure TImgBlockStore.setDevBSize(size: Int64);
var newOffset : Int64;
begin
  if (size<=0) or (not Assigned(FDataSource)) or
     (FDataSource.rescueStatus.devicesize mod size <> 0) then begin
    MainForm.DbgLog.Lines.Add('Cannot set device sector size = ' + inttostr(size));
    exit;
  end;
  if (size <> FDevBSize) then begin
    newOffset := (FBOffset*FDevBSize) div size;
    FBCount := ((FBOffset+FBCount)*FDevBSize) div size - newOffset;
    FBOffset := newOffset;
    FDevBSize := size;
  end;
end;

// returns a bit mask for each known block type
function TImgBlockStore.BlockStatusToMask(status: char) : Longword;
begin
  case status of
    '?': result := MASK_NON_TRIED;
    '*': result := MASK_NON_TRIMMED;
    '/': result := MASK_NON_SCRAPED;
    '-': result := MASK_BAD_SECT;
    '+': result := MASK_FINISHED;
  else
    result := 0;
  end;
end;

// zoom in or out by some factor, centered around a given imgBlock
// factor <1 zooms in, factor >1 zooms out.
Procedure TImgBlockStore.zoom(block: Longint; factor: double);
var
  zoomPos, zoomOffset, zoomLength, deviceSectors : Int64; // all values in sectors
begin
  if not FDataSource.hasFile() then exit;
  zoomPos:=getOffsetFromBlock(block) div FdevBSize;
  deviceSectors:=FDataSource.rescueStatus.devicesize div FdevBSize;
  if factor < 1.0 then begin // zoom in
    // if already at 1 sector per block -> dont zoom further
    if FBCount <= FnumBlocks then exit;
    zoomLength:=Max(FnumBlocks, Min(FBCount-FnumBlocks, FBCount * Round(128*factor) div 128));
  end else begin // zoom out
    zoomLength:=Min(deviceSectors, FBCount * Round(128*factor) div 128);
  end;
  //MainForm.AppLog.Lines.Add('zooming to '+inttostr(zoomlength)+'sectors in '+inttostr(FnumBlocks)+' blocks');
  FBCount:=zoomLength;
  FBOffset:=Max(0, Min(zoomPos-(zoomLength div 2), deviceSectors-zoomLength));
end;

end.

